from abc import ABC, abstractmethod
from itertools import product
import torch
from pathlib import Path
import numpy as np
from typing import List, Dict, Union


def records_to_list(list_of_dicts: Union[List[Dict[str, int]], Dict[str, int]]):
    """Transform a list of dictionaries to a dictionary of lists.
    If list_of_dicts is a dictionary, it will simply make a list of each values.
    From: https://stackoverflow.com/questions/5558418/list-of-dicts-to-from-dict-of-lists

    Args:
        list_of_dicts (list[dict[str, int]] | dict[str, int]): List of dictionaries, assuming each dictionary has the same keys.

    Returns:
        dict[list]: Dictionary of lists.
    """
    if not isinstance(list_of_dicts, list):
        list_of_dicts = [list_of_dicts]
    return {k: [dic[k] for dic in list_of_dicts] for k in list_of_dicts[0]}


def chunk(L, n):
    """
    Partition L into n chunks using every item in L and
    such that the resulting chunks differ in size by at
    most one element.

    >>> L = ['a', 'b', 'c', 'd']
    ['a', 'b', 'c', 'd']
    >>> chunk(L, 2)
    [['a', 'b'], ['c', 'd']]
    >>> chunk(L, 3)
    [['a'], ['b', 'c'], ['d']]
    >>> chunk(L, 4)
    [['a'], ['b'], ['c'], ['d']]
    >>> chunk(L, 5)
    [['a'], ['b'], [], ['c'], ['d']]
    """
    size = len(L) / float(n)
    I = lambda i: int(round(i))
    return [L[I(size * i) : I(size * (i + 1))] for i in range(n)]


class AbstractCheckpoints(ABC):
    """Abstract class for iterating over model checkpoints"""

    def __init__(self, device: str = "cpu"):
        self.low_cpu_mem_usage = True if device == "cpu" else False

        self._device = device
        if device == "cpu":
            self.device = torch.device("cpu")
        elif device == "cuda":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        elif device == "mps":
            self.device = torch.device(
                "mps" if torch.backends.mps.is_available() else "cpu"
            )

    def split(self, n):
        """Convenience function for splitting checkpoints for e.g. parallel computing.
        If n > m (number of total checkpoints), it will return a list of m checkpoints objects instead.
        """
        total_length = len(self)
        chunk_indices = np.array_split(range(total_length), n)

        ckpts = []
        for ci in chunk_indices:
            if len(ci) > 0:
                start = ci[0].item()
                end = ci[-1].item()
                cfg = self.config
                cfg.update(
                    {
                        k: set(v)
                        for k, v in records_to_list(
                            self.checkpoints[start : end + 1]
                        ).items()
                    }
                )
                # ckpts.append(self.__class__(seeds=set(cfg["seed"]), steps=set(cfg["step"])))
                ckpts.append(self.__class__(**cfg))
        return ckpts

    @property
    @abstractmethod
    def name():
        pass

    @property
    @abstractmethod
    def checkpoints(self):
        pass

    @staticmethod
    @abstractmethod
    def last_step():
        pass

    @abstractmethod
    def get_checkpoint(self):
        pass

    @classmethod
    def final_checkpoints(cls, **kwargs):
        return cls(step=[cls.last_step()], **kwargs)

    @abstractmethod
    def __len__(self):
        pass

    def __getitem__(self, index):
        cfg = self.checkpoints[index]
        ckpt = self.get_checkpoint(**cfg)
        print(ckpt)
        return ckpt

    def __iter__(self):
        for cfg in self.checkpoints:
            ckpt = self.get_checkpoint(**cfg)
            yield ckpt


class Checkpoint:
    """Convenience class for representing a checkpoint.
    Each checkpoint should at least have a model and model_name.
    """

    def __init__(self, model, model_name, tokenizer=None, **kwargs):
        self._model = model
        self._tokenizer = tokenizer
        self._config = {"model_name": model_name}
        self._config.update(kwargs)

    @property
    def config(self):
        return self._config

    @property
    def model(self):
        return self._model

    @property
    def tokenizer(self):
        if not self._tokenizer:
            raise ValueError("This checkpoint has no corresponding tokenizer.")
        return self._tokenizer



import torch
import pytest

_available_devices = ["cpu"]
if torch.cuda.is_available():
    _available_devices.append("cuda:0")
if torch.backends.mps.is_available():
    _available_devices.append("mps")


def multi_device(test_method):
    """
    Adapted from AllenNLP.

    Decorator that provides an argument `device` of type `str` to a test function.

    If you have a CUDA capable GPU available, device will be "cuda:0", otherwise the device will
    be "cpu".

    !!! Note
        If you have a CUDA capable GPU available, but you want to run the test using CPU only,
        just set the environment variable "CUDA_CAPABLE_DEVICES=''" before running pytest.
    """
    return pytest.mark.parametrize("device", _available_devices)(test_method)



from lm_checkpoints import AbstractCheckpoints, Checkpoint
from itertools import product
from transformers import AutoTokenizer, GPTNeoXForCausalLM
from typing import List, Dict


class PythiaCheckpoints(AbstractCheckpoints):
    """Class for iterating over Pythia checkpoints"""

    def __init__(
        self,
        size: int = 70,
        step: List[int] = None,
        seed: List[int] = None,
        deduped: bool = False,
        **kwargs,
    ) -> None:
        """Initialize the PythiaCheckpoints.

        Args:
            size (int): Model size. Defaults to 70.
            step (List[int], optional): List of steps to consider, uses all available steps if not specified.
            seed (List[int], optional): List of seeds to consider, uses all available seeds if not specified.
            deduped (bool, optional): Specifies whether to use the deduped version of Pythia. Defaults to False.
        """
        super().__init__(**kwargs)

        self.deduped = deduped
        self.size = size
        if deduped:
            raise NotImplementedError

        self._seeds = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        if seed:
            assert set(seed).issubset(set(self._seeds))
            self.seeds = seed
        else:
            self.seeds = self._seeds

        self._steps = [0, 1, 2, 4, 8, 16, 32, 64, 128, 256, 512] + list(
            range(1000, 144000, 1000)
        )
        if step:
            assert set(step).issubset(set(self._steps))
            self.steps = step
        else:
            self.steps = self._steps

    @property
    def name(self) -> str:
        return f"Pythia {self.size}m"

    @staticmethod
    def last_step() -> int:
        """Last step of training."""
        return 143000

    @property
    def config(self) -> dict:
        """Returns a dictionary for re-initializing this checkpoints class.

        Returns:
            dict: Configuration of this checkpoints object.
        """
        return {"size": self.size, "deduped": self.deduped}

    def get_model_name(self, seed: int) -> str:
        """Get the name for loading from the HF hub.

        Args:
            seed (int): Model seed.

        Returns:
            str: Name of the checkpoint on HF.
        """
        if seed == 0:
            return f"EleutherAI/pythia-{self.size}m"
        else:
            return f"EleutherAI/pythia-{self.size}m-seed{seed}"

    @property
    def checkpoints(self) -> List[Dict[str, int]]:
        """Returns all step and seed combinations that make up the checkpoints.

        Returns:
            list[dict[str, int]]: List of dicts (seed, step) describing each checkpoint.
        """
        return list(
            {"seed": p[0], "step": p[1]} for p in product(self.seeds, self.steps)
        )

    def __len__(self):
        return len(self.seeds) * len(self.steps)

    def get_checkpoint(self, seed, step) -> Checkpoint:
        model_name = self.get_model_name(seed)
        tokenizer = AutoTokenizer.from_pretrained(
            model_name,
            revision=f"step{step}",
        )

        model = GPTNeoXForCausalLM.from_pretrained(
            model_name,
            revision=f"step{step}",
            low_cpu_mem_usage=self.low_cpu_mem_usage,
        )
        model.eval()
        model = model.to(self.device)

        return Checkpoint(
            model, tokenizer=tokenizer, model_name=model_name, seed=seed, step=step
        )



from lm_checkpoints import AbstractCheckpoints, Checkpoint
from transformers import AutoConfig, AutoTokenizer, AutoModelForMaskedLM
from itertools import product
from typing import List, Dict


class MultiBERTCheckpoints(AbstractCheckpoints):
    """Class for iterating over MultiBERT checkpoints"""

    def __init__(self, step=None, seed=None, **kwargs):
        """Initialize the MultiBERTCheckpoints.

        Args:
            step (List[int], optional): List of steps to consider, uses all available steps if not specified.
            seed (List[int], optional): List of seeds to consider, uses all available seeds if not specified.
        """
        super().__init__(**kwargs)

        self._seeds = [0, 1, 2, 3, 4]
        if seed:
            assert set(seed).issubset(set(self._seeds))
            self.seeds = seed
        else:
            self.seeds = self._seeds

        self._steps = [
            0,
            20,
            40,
            60,
            80,
            100,
            120,
            140,
            160,
            180,
            200,
            300,
            400,
            500,
            600,
            700,
            800,
            900,
            1000,
            1100,
            1200,
            1300,
            1400,
            1500,
            1600,
            1700,
            1800,
            1900,
            2000,
        ]
        if step:
            assert set(step).issubset(set(self._steps))
            self.steps = step
        else:
            self.steps = self._steps

    @property
    def name(self) -> str:
        return f"MultiBERTs"

    @staticmethod
    def last_step() -> int:
        """Last step of training."""
        return 2000

    @property
    def config(self) -> dict:
        """Returns a dictionary for re-initializing this checkpoints class.

        Returns:
            dict: Configuration of this checkpoints object.
        """
        return {}

    def get_model_name(self, step: int, seed: int) -> str:
        """Get the name for loading from the HF hub.

        Args:
            step (int): Checkpoint step.
            seed (int): Model seed.

        Returns:
            str: Name of the checkpoint on HF.
        """
        return f"google/multiberts-seed_{seed}-step_{step}k"

    @property
    def checkpoints(self) -> List[Dict[str, int]]:
        """Returns all step and seed combinations that make up the checkpoints.

        Returns:
            list[dict[str, int]]: List of dicts (seed, step) describing each checkpoint.
        """
        return list(
            {"seed": p[0], "step": p[1]} for p in product(self.seeds, self.steps)
        )

    def __len__(self):
        return len(self.seeds) * len(self.steps)

    def get_checkpoint(self, seed, step) -> Checkpoint:
        model_name = self.get_model_name(step, seed)
        config = AutoConfig.from_pretrained(model_name)

        # Necessary here?
        config.output_scores = True

        tokenizer = AutoTokenizer.from_pretrained(
            model_name,
            device=self.device,
        )

        model = AutoModelForMaskedLM.from_pretrained(
            model_name,
            config=config,
            low_cpu_mem_usage=self.low_cpu_mem_usage,  # https://huggingface.co/docs/transformers/main_classes/model#large-model-loading
        )
        model.eval()
        model = model.to(self.device)

        return Checkpoint(
            model, tokenizer=tokenizer, model_name=model_name, seed=seed, step=step
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lm_checkpoints', 'lm_checkpoints.testing']

package_data = \
{'': ['*']}

install_requires = \
['accelerate>=0.24.1,<0.25.0',
 'torch>=2.0.0,!=2.0.1,!=2.1.0',
 'transformers>=4.35.0,<5.0.0']

setup_kwargs = {
    'name': 'lm-checkpoints',
    'version': '0.1.10',
    'description': 'Simple library for loading checkpoints of language models.',
    'long_description': '# lm-checkpoints\nSimple library for dealing with language model checkpoints.\n\nInstall using `pip install lm-checkpoints`.\n\n## Example\nSay you want to compute some metrics for all model checkpoints of Pythia 160m, but only seed 0.\n\n```\nfrom lm_checkpoints import PythiaCheckpoints\n\nfor ckpt in PythiaCheckpoints(size=160,seed=[0]):\n    # Do something with ckpt.model or ckpt.tokenizer\n```\n\nOr if you only want to load steps `0, 1, 2, 4, 8, 16` for all available seeds:\n```\nfrom lm_checkpoints import PythiaCheckpoints\n\nfor ckpt in PythiaCheckpoints(size=160,step=[0, 1, 2, 4, 8, 16]):\n    # Do something with ckpt.model or ckpt.tokenizer\n```\n\nAlternatively, you may want to load all final checkpoints of MultiBERTs:\n```\nfrom lm_checkpoints import MultiBERTCheckpoints\n\nfor ckpt in MultiBERTCheckpoints.final_checkpoints():\n    # Do something with ckpt.model or ckpt.tokenizer\n```',
    'author': 'Oskar van der Wal',
    'author_email': 'odw@duck.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
